/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef __AIDGE_CORE_OPERATOR_LEAKYRELU_H__
#define __AIDGE_CORE_OPERATOR_LEAKYRELU_H__

#include <vector>
#include <memory>

#include "aidge/utils/Parameter.hpp"
#include "aidge/utils/Registrar.hpp"
#include "aidge/operator/Operator.hpp"
#include "aidge/backend/OperatorImpl.hpp"
#include "aidge/data/Tensor.hpp"
#include "aidge/data/Data.hpp"
#include "aidge/graph/Node.hpp"
#include "aidge/utils/Types.h"

namespace Aidge {
enum class LeakyReLUParam {
    NegativeSlope
};

class LeakyReLU_Op : public Operator,
    public Registrable<LeakyReLU_Op, std::string, std::unique_ptr<OperatorImpl>(const LeakyReLU_Op&)>,
    public Parameterizable<LeakyReLUParam, float> {
public:
    // FIXME: change accessibility
    std::shared_ptr<Tensor> mInput = std::make_shared<Tensor>();
    const std::shared_ptr<Tensor> mOutput = std::make_shared<Tensor>();

public:
    static constexpr const char* Type = "LeakyReLU";

    LeakyReLU_Op() = delete;

    using Parameterizable_ = Parameterizable<LeakyReLUParam, float>;
    template <LeakyReLUParam e> using param = typename Parameterizable_::template param<e>;

    LeakyReLU_Op(float negativeSlope)
            : Operator(Type),
            Parameterizable_(
                param<LeakyReLUParam::NegativeSlope>(negativeSlope))
    {
        setDatatype(DataType::Float32);
    }

    void associateInput(__attribute__((unused)) const IOIndex_t inputIdx, std::shared_ptr<Data> data) override final {
        assert(inputIdx == 0 && "operator supports only 1 input");
        assert(strcmp(data->type(), Tensor::Type)==0 && "input data must be of Tensor type");
        mInput = std::dynamic_pointer_cast<Tensor>(data);
    }

    void computeOutputDims() override final {
        if (!mInput->empty())
            mOutput->resize(mInput->dims());
    }

    bool outputDimsForwarded() const override final {
        return !(mOutput->empty());
    }


    inline Tensor& input(__attribute__((unused)) const IOIndex_t inputIdx) const override final { return *(mInput.get()); }
    inline Tensor& output(__attribute__((unused)) const IOIndex_t outputIdx) const override final { return *(mOutput.get()); }


    inline std::shared_ptr<Tensor> getInput(__attribute__((unused)) const IOIndex_t inputIdx) const override final { 
        assert((inputIdx == 0) && "LeakyReLU Operator has only 1 input");
        return mInput;
    }
    inline std::shared_ptr<Tensor> getOutput(__attribute__((unused)) const IOIndex_t outputIdx) const override final {
        assert((outputIdx == 0) && "LeakyReLU Operator has only 1 output");
        return mOutput;
    }


    std::shared_ptr<Data> getRawInput(__attribute__((unused)) const IOIndex_t inputIdx) const override final {
        assert(inputIdx == 0 && "operator supports only 1 input");
        return std::static_pointer_cast<Data>(mInput);
    }
    std::shared_ptr<Data> getRawOutput(__attribute__((unused)) const IOIndex_t outputIdx) const override final {
        assert(outputIdx == 0 && "operator supports only 1 output");
        return mOutput;
    }


    void setBackend(const std::string& name) {
        mImpl = Registrar<LeakyReLU_Op>::create(name)(*this);
        mOutput->setBackend(name);

        // FIXME: temporary workaround
        mInput->setBackend(name);
    }
    void setDatatype(const DataType& datatype) {
        mOutput->setDatatype(datatype);

        // FIXME: temporary workaround
        mInput->setDatatype(datatype);
    }

    inline IOIndex_t nbInputs() const noexcept override final { return 1; }
    inline IOIndex_t nbDataInputs() const noexcept override final { return 1; }
    inline IOIndex_t nbOutputs() const noexcept override final { return 1; }
};

inline std::shared_ptr<Node> LeakyReLU(float negativeSlope = 0.0f, const char* name = nullptr) {
    // FIXME: properly handle default w&b initialization in every cases
    return std::make_shared<Node>(std::make_shared<LeakyReLU_Op>(negativeSlope), name);
}
}

namespace {
template <>
const char* const EnumStrings<Aidge::LeakyReLUParam>::data[]
    = {"NegativeSlope"};
}

#endif /* __AIDGE_CORE_OPERATOR_RELU_H__ */
