/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef __AIDGE_CORE_OPERATOR_MATMUL_H__
#define __AIDGE_CORE_OPERATOR_MATMUL_H__

#include <array>
#include <cmath>
#include <numeric>
#include <memory>
#include <vector>

#include "aidge/utils/Types.h"
#include "aidge/data/Tensor.hpp"
#include "aidge/graph/Node.hpp"
#include "aidge/operator/Operator.hpp"
#include "aidge/operator/Producer.hpp"
#include "aidge/utils/Parameter.hpp"
#include "aidge/utils/Registrar.hpp"

namespace Aidge {
enum class MatmulParam { OutChannels };

class Matmul_Op : public Operator,
              public Registrable<Matmul_Op,
                                 std::string,
                                 std::unique_ptr<OperatorImpl>(const Matmul_Op &)>,
              public Parameterizable<MatmulParam, DimSize_t> {
public:
    std::array<std::shared_ptr<Tensor>, 2> mInputs = {std::make_shared<Tensor>(), std::make_shared<Tensor>()};
    const std::shared_ptr<Tensor> mOutput = std::make_shared<Tensor>();

public:
    static constexpr const char* Type = "Matmul";

    Matmul_Op() = delete;

    using Parameterizable_ = Parameterizable<MatmulParam, DimSize_t>;
    template <MatmulParam e> using param = typename Parameterizable_::template param<e>;

    Matmul_Op(DimSize_t out_channels)
            : Operator(Type),
            Parameterizable_(
                param<MatmulParam::OutChannels>(out_channels)),
            mOutput(std::make_shared<Tensor>())
    {
        setDatatype(DataType::Float32);
    }

    void associateInput(__attribute__((unused)) const IOIndex_t inputIdx, std::shared_ptr<Data> data) override final {
        assert(inputIdx < 2 && "operators supports only 2 inputs");
        assert(strcmp(data->type(), Tensor::Type)==0 && "input data must be of Tensor type");
        mInputs[inputIdx] = std::dynamic_pointer_cast<Tensor>(data);
    }

    void computeOutputDims() override final {
        if (!mInputs[0]->empty()) {
            // <in_features**, out_channels>
            std::array<DimSize_t, 2> weightDims = {static_cast<DimSize_t>(mInputs[0]->size()), this->template get<MatmulParam::OutChannels>()};
            // <out_channels, batch>
            std::array<DimSize_t, 1> outputDims = {this->template get<MatmulParam::OutChannels>()};
            
            mInputs[1]->resize(weightDims);
            mOutput->resize(outputDims);
        }
    }

    bool outputDimsForwarded() const override final {
        return !(mOutput->empty());
    }


    inline Tensor& input(const IOIndex_t inputIdx) const override final {
        assert(inputIdx < 2 && "operators supports only 2 inputs");
        return *(mInputs[inputIdx].get()); }
    inline Tensor& output(__attribute__((unused)) const IOIndex_t outputIdx) const override final { return *(mOutput.get()); }


    inline std::shared_ptr<Tensor> getInput(const IOIndex_t inputIdx) const override final {
        assert(inputIdx < 2 && "MatMul Operators has 2 inputs");
        return mInputs[inputIdx];
    }
    inline std::shared_ptr<Tensor> getOutput(__attribute__((unused)) const IOIndex_t outputIdx) const override final {
        assert((outputIdx == 0) && "MatMul Operators has 1 output");
        return mOutput;
    }


    std::shared_ptr<Data> getRawInput(const IOIndex_t inputIdx) const override final {
        assert(inputIdx < 2 && "operators supports only 2 inputs");
        return std::static_pointer_cast<Data>(mInputs[inputIdx]);
    }
    std::shared_ptr<Data> getRawOutput(__attribute__((unused)) const IOIndex_t outputIdx) const override final {
        assert(outputIdx == 0 && "operator supports only 1 output");
        return std::static_pointer_cast<Data>(mOutput);
    }


    void setBackend(const std::string& name) {
        mImpl = Registrar<Matmul_Op>::create(name)(*this);
        mOutput->setBackend(name);

        // FIXME: temporary workaround
        mInputs[0]->setBackend(name);
        mInputs[1]->setBackend(name);
    }

    void setDatatype(const DataType& datatype) {
        mOutput->setDatatype(datatype);

        // FIXME: temporary workaround
        mInputs[0]->setDatatype(datatype);
        mInputs[1]->setDatatype(datatype);
    }


    inline IOIndex_t nbInputs() const noexcept override final { return 2; }
    inline IOIndex_t nbDataInputs() const noexcept override final { return 1; }
    inline IOIndex_t nbOutputs() const noexcept override final { return 1; }
};

inline std::shared_ptr<Node> Matmul(DimSize_t out_channels, const char* name = nullptr) {
    // FIXME: properly handle default w&b initialization in every cases
    auto matmul = std::make_shared<Node>(std::make_shared<Matmul_Op>(out_channels), name);
    addProducer(matmul, 1, {1, out_channels}, "w");
    return matmul;
}
} // namespace Aidge

namespace {
template <>
const char *const EnumStrings<Aidge::MatmulParam>::data[] = {"OutChannels"};
}

#endif /* __AIDGE_CORE_OPERATOR__MATMUL_H__ */
